import os

import pytest
from greyhorse_process.local import AsyncLocalConnection


@pytest.mark.asyncio
async def test_async_local_run(local_root_password) -> None:
    conn = AsyncLocalConnection(sudo_password=local_root_password)

    async with conn.session() as session:
        res = await session.run('echo "123"', shell=True, as_bytes=True)
        assert res.stdout == b'123\n'

        res = await session.run('echo "123"', shell=True)
        assert res.stdout == '123'

        res = await session.run('pwd', shell=False)
        assert res.stdout == os.getcwd()

        res = await session.sudo('whoami', shell=True)
        assert res.stdout == 'root'


@pytest.mark.asyncio
async def test_async_local_interactive(local_root_password) -> None:
    conn = AsyncLocalConnection(sudo_password=local_root_password)

    async with conn.session() as session:
        async with session.create_process(
            f'{os.getcwd()}/tests/echo.sh', as_bytes=True
        ) as proc:
            line1 = await proc.read_line()
            await proc.write_line(b'123')
            line2 = await proc.read_line()
            await proc.write_line(b'456')
            line3 = await proc.read_line()

        assert line1 == b'hello'
        assert line2 == b'hello 123'
        assert line3 == b'bye 456'

        async with session.create_process(f'{os.getcwd()}/tests/echo.sh') as proc:
            line1 = await proc.read_line()
            await proc.write_line('123')
            line2 = await proc.read_line()
            await proc.write_line('456')
            line3 = await proc.read_line()

        assert line1 == 'hello'
        assert line2 == 'hello 123'
        assert line3 == 'bye 456'



import asyncio
import shlex
from collections.abc import Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import override

from greyhorse.logging import logger

from .abc import AsyncConnection, AsyncSession, CompletedProcess
from .adapters import AsyncProcessAdapter


class AsyncLocalSession(AsyncSession):
    def __init__(self, sudo_password: str | None = None) -> None:
        self._sudo_password = sudo_password

    @override
    @asynccontextmanager
    async def create_process(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> Callable[..., AbstractAsyncContextManager[AsyncProcessAdapter]]:
        command = shlex.split(command, comments=True)

        if shell:
            entrypoint = asyncio.create_subprocess_shell
        else:
            entrypoint = asyncio.create_subprocess_exec

        if sudo:
            if self._sudo_password is None:
                logger.warning('Sudo used without the password')
            command = ['sudo', '-S', *command]

        if shell:
            command = [' '.join(command)]

        process = await entrypoint(
            *command,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        proc = AsyncProcessAdapter(process, encoding=None if as_bytes else 'utf-8')
        if sudo and self._sudo_password:
            await proc.write_line(self._sudo_password)
            await proc.read_line(from_stderr=True)
        yield proc

    @override
    async def run(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> CompletedProcess:
        encoding = None if as_bytes else 'utf-8'
        input = input.encode('utf-8') if isinstance(input, str) else input

        async with self.create_process(command, shell, sudo=sudo, as_bytes=as_bytes) as process:
            logger.debug(f'Local proc started: {command!r} pid: {process.wrapped.pid}')
            stdout, stderr = await process.wrapped.communicate(input=input)

        if process.returncode == 0:
            logger.debug(f'Local proc done: {command!r} pid: {process.wrapped.pid}')
        else:
            logger.debug(
                f'Local proc failed: {command!r} pid: {process.wrapped.pid} code: {process.returncode}'
            )

        if encoding:
            stdout = stdout.decode(encoding).rstrip() if stdout else ''
            stderr = stderr.decode(encoding).rstrip() if stderr else ''
        else:
            stdout, stderr = stdout if stdout else b'', stderr if stderr else b''

        return CompletedProcess(
            command=command, returncode=process.returncode, stdout=stdout, stderr=stderr
        )


class AsyncLocalConnection(AsyncConnection):
    def __init__(self, sudo_password: str | None = None) -> None:
        self._sudo_password = sudo_password

    @override
    @asynccontextmanager
    async def session(self) -> AbstractAsyncContextManager[AsyncLocalSession]:
        yield AsyncLocalSession(self._sudo_password)



from collections.abc import AsyncIterator, Iterator
from typing import AnyStr


class SyncProcessAdapter:
    def __init__(self, process, encoding: str | None = None) -> None:
        self._proc = process
        self._encoding = encoding
        self.stdin = process.stdin
        self.stdout = process.stdout
        self.stderr = process.stderr

    def read_line(self, from_stderr: bool = False, rstrip: bool = True) -> bytes | str:
        from_ = self._proc.stderr if from_stderr else self._proc.stdout
        data = from_.readline()
        if self._encoding and isinstance(data, bytes):
            data = data.decode(self._encoding)
        if rstrip:
            data = data.rstrip()
        return data

    def write_line(self, data: str | bytes) -> None:
        if self._encoding and isinstance(data, str):
            data = data.encode(self._encoding)
        data = data + (b'\n' if isinstance(data, bytes) else '\n')
        self._proc.stdin.write(data)
        self._proc.stdin.drain()

    def write_eof(self) -> None:
        self._proc.stdin.write_eof()
        self._proc.stdin.drain()

    class _ReadlineIter:
        def __init__(self, proc: 'SyncProcessAdapter', from_stderr: bool, rstrip: bool) -> None:
            self._proc = proc
            self._from_stderr = from_stderr
            self._rstrip = rstrip

        def __iter__(self) -> Iterator[AnyStr]:
            from_ = self._proc.stderr if self._from_stderr else self._proc.stdout
            while not from_.at_eof():
                yield self._proc.read_line(self._from_stderr, self._rstrip)

    def readline_iter(self, from_stderr: bool = False, rstrip: bool = True):
        return SyncProcessAdapter._ReadlineIter(self, from_stderr, rstrip)

    def wait(self):
        return self._proc.wait()

    @property
    def returncode(self):
        return self._proc.returncode

    @property
    def wrapped(self):
        return self._proc


class AsyncProcessAdapter:
    def __init__(self, process, encoding: str | None = None) -> None:
        self._proc = process
        self._encoding = encoding
        self.stdin = process.stdin
        self.stdout = process.stdout
        self.stderr = process.stderr

    async def read_line(self, from_stderr: bool = False, rstrip: bool = True) -> bytes | str:
        from_ = self._proc.stderr if from_stderr else self._proc.stdout
        data = await from_.readline()
        if self._encoding and isinstance(data, bytes):
            data = data.decode(self._encoding)
        if rstrip:
            data = data.rstrip()
        return data

    async def write_line(self, data: str | bytes) -> None:
        if self._encoding and isinstance(data, str):
            data = data.encode(self._encoding)
        data = data + (b'\n' if isinstance(data, bytes) else '\n')
        self._proc.stdin.write(data)
        await self._proc.stdin.drain()

    async def write_eof(self) -> None:
        self._proc.stdin.write_eof()
        await self._proc.stdin.drain()

    class _ReadlineIter:
        def __init__(
            self, proc: 'AsyncProcessAdapter', from_stderr: bool, rstrip: bool
        ) -> None:
            self._proc = proc
            self._from_stderr = from_stderr
            self._rstrip = rstrip

        async def __aiter__(self) -> AsyncIterator[AnyStr]:
            from_ = self._proc.stderr if self._from_stderr else self._proc.stdout
            while not from_.at_eof():
                yield await self._proc.read_line(self._from_stderr, self._rstrip)

    def readline_iter(self, from_stderr: bool = False, rstrip: bool = True):
        return AsyncProcessAdapter._ReadlineIter(self, from_stderr, rstrip)

    async def wait(self):
        return await self._proc.wait()

    @property
    def returncode(self):
        return self._proc.returncode

    @property
    def wrapped(self):
        return self._proc



import os

import pytest


@pytest.fixture
def local_root_password():
    return os.getenv('LOCAL_ROOT_PASSWORD')


@pytest.fixture
def ssh_host():
    return os.getenv('SSH_HOST')


@pytest.fixture
def ssh_username():
    return os.getenv('SSH_USERNAME')


@pytest.fixture
def ssh_root_password():
    return os.getenv('SSH_ROOT_PASSWORD')



from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Callable, Generator
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    asynccontextmanager,
    contextmanager,
)
from dataclasses import dataclass

from .adapters import AsyncProcessAdapter, SyncProcessAdapter


@dataclass(slots=True, frozen=True)
class CompletedProcess:
    command: str
    returncode: int
    stdout: str | bytes
    stderr: str | bytes

    @property
    def ok(self) -> bool:
        return self.returncode == 0


class SyncSession(ABC):
    @contextmanager
    @abstractmethod
    def create_process(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> Callable[..., AbstractContextManager[SyncProcessAdapter]]: ...

    @abstractmethod
    def run(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> CompletedProcess: ...

    def sudo(
        self,
        command: str,
        shell: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> CompletedProcess:
        return self.run(command, shell=shell, sudo=True, as_bytes=as_bytes, input=input)


class AsyncSession(ABC):
    @asynccontextmanager
    @abstractmethod
    async def create_process(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> Callable[..., AbstractAsyncContextManager[AsyncProcessAdapter]]: ...

    @abstractmethod
    async def run(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> CompletedProcess: ...

    async def sudo(
        self,
        command: str,
        shell: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> CompletedProcess:
        return await self.run(command, shell=shell, sudo=True, as_bytes=as_bytes, input=input)


class SyncConnection(ABC):
    @abstractmethod
    def session(self) -> Generator[AbstractContextManager[SyncSession], None, None]: ...


class AsyncConnection(ABC):
    @abstractmethod
    async def session(
        self,
    ) -> AsyncGenerator[AbstractAsyncContextManager[AsyncSession], None]: ...



import pytest
from greyhorse_process.ssh import AsyncSshConnection


@pytest.mark.asyncio
async def test_async_ssh_run(ssh_host, ssh_username, ssh_root_password) -> None:
    conn = AsyncSshConnection(ssh_host, username=ssh_username, sudo_password=ssh_root_password)

    async with conn.session() as session:
        res = await session.run('echo "123"', shell=True, as_bytes=True)
        assert res.stdout == b'123\n'

        res = await session.run('echo "123"', shell=True)
        assert res.stdout == '123'

        res = await session.run('pwd', shell=False)
        assert res.stdout == '/root'

        res = await session.sudo('whoami', shell=True)
        assert res.stdout == 'root'


@pytest.mark.asyncio
async def test_async_ssh_interactive(ssh_host, ssh_username) -> None:
    conn = AsyncSshConnection(ssh_host, username=ssh_username)

    async with conn.session() as session:
        async with session.create_process('/root/echo.sh', as_bytes=True) as proc:
            line1 = await proc.read_line()
            await proc.write_line(b'123')
            line2 = await proc.read_line()
            await proc.write_line(b'456')
            line3 = await proc.read_line()

        assert line1 == b'hello'
        assert line2 == b'hello 123'
        assert line3 == b'bye 456'

        async with session.create_process('/root/echo.sh') as proc:
            line1 = await proc.read_line()
            await proc.write_line('123')
            line2 = await proc.read_line()
            await proc.write_line('456')
            line3 = await proc.read_line()

        assert line1 == 'hello'
        assert line2 == 'hello 123'
        assert line3 == 'bye 456'



import shlex
from collections.abc import Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from datetime import timedelta
from typing import override

import asyncssh
from asyncssh import SSHClientConnection, SSHClientConnectionOptions
from greyhorse.logging import logger

from .abc import AsyncConnection, AsyncSession, CompletedProcess
from .adapters import AsyncProcessAdapter


class AsyncSshSession(AsyncSession):
    def __init__(
        self,
        connection: SSHClientConnection,
        server_string: str,
        sudo_password: str | None = None,
    ) -> None:
        self._connection = connection
        self._sudo_password = sudo_password
        self._server_string = server_string

    @property
    def raw_connection(self) -> SSHClientConnection:
        return self._connection

    @override
    @asynccontextmanager
    async def create_process(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> Callable[..., AbstractAsyncContextManager[AsyncProcessAdapter]]:
        use_terminal = False
        input = input.encode('utf-8') if isinstance(input, str) else input
        command = shlex.split(command, comments=True)

        if sudo:
            if self._sudo_password is None:
                logger.warning('Sudo used without the password')
            use_terminal = True
            command = ['sudo', '-S', *command]

        if shell:
            command = [' '.join(command)]

        async with self._connection.create_process(
            *command, encoding=None, input=input, term_type='term' if use_terminal else None
        ) as process:
            proc = AsyncProcessAdapter(process, encoding=None if as_bytes else 'utf-8')
            if sudo and self._sudo_password:
                await proc.write_line(self._sudo_password)
                await proc.read_line()
            yield proc

    @override
    async def run(
        self,
        command: str,
        shell: bool = False,
        sudo: bool = False,
        as_bytes: bool = False,
        input: str | bytes | None = None,
    ) -> CompletedProcess:
        encoding = None if as_bytes else 'utf-8'

        async with self.create_process(
            command, shell, sudo=sudo, as_bytes=as_bytes, input=input
        ) as process:
            logger.debug(f'SSH proc started: {command!r} server: "{self._server_string}"')
            process = await process.wait()

        stdout, stderr = process.stdout, process.stderr

        if process.returncode == 0:
            logger.debug(f'SSH proc done: {command!r} server: "{self._server_string}"')
        else:
            logger.debug(
                f'SSH proc failed: {command!r} server: "{self._server_string}" '
                f'code: {process.returncode}'
            )

        if encoding:
            if isinstance(stdout, bytes):
                stdout = stdout.decode(encoding)
            elif not stdout:
                stdout = ''
            if isinstance(stderr, bytes):
                stderr = stderr.decode(encoding)
            elif not stderr:
                stderr = ''

            stdout, stderr = stdout.rstrip(), stderr.rstrip()
        else:
            stdout, stderr = stdout if stdout else b'', stderr if stderr else b''

        return CompletedProcess(
            command=command, returncode=process.returncode, stdout=stdout, stderr=stderr
        )


class AsyncSshConnection(AsyncConnection):
    def __init__(
        self,
        host: str,
        port: int = 22,
        username: str | None = None,
        password: str | None = None,
        sudo_password: str | None = None,
        connect_timeout: timedelta | None = None,
    ) -> None:
        self._host = host
        self._port = port or 22
        self._username = username
        self._password = password
        self._sudo_password = sudo_password
        self._connect_timeout = connect_timeout

    @override
    @asynccontextmanager
    async def session(self) -> AbstractAsyncContextManager[AsyncSshSession]:
        options = SSHClientConnectionOptions(
            username=self._username,
            password=self._password,
            known_hosts=None,
            connect_timeout=self._connect_timeout.total_seconds()
            if self._connect_timeout
            else None,
        )

        async with asyncssh.connect(
            host=self._host, port=self._port or (), options=options
        ) as connection:  # type: SSHClientConnection
            yield AsyncSshSession(
                connection=connection,
                server_string=f'{self._username}@{self._host}:{self._port}',
                sudo_password=self._sudo_password,
            )
